
listFunc = lambda lst, off=0: lambda i: lst[i+off]
"""Represent a list *lst* through a function f, such that
f(x) = *lst* [ x + *off* ]."""

def isList(l):
    """Check, if l is a list type or not."""
    try:
        len(l)
        return True
    except:
        return False


_tail = lambda l: l[-1]

# FIXME: use zip(v), zip(*v)
class Format:
    """Data format conversion and tools.
    
    There are three kind of items:
    
    - Feature
    - Label
    - Weight
    
    The items are always ordered Feature, Label, Weight but items may be tied
    together using a tuple.
    
    - *packed*:         [(Feature, Label, Weight)]
    - *hipack*:         [(Feature, Label)], [Weights]
    - *lopack*:         [Features], [(Label, Weight)]
    - *seperated*:      [Features], [Labels], [Weights]
    
    Note that the items are lists but when two items are packed in a
    tuple, the resulting list will contain tuples of the items.
    
    This class uses subclasses for conversion of one type into another.
    Example:
    
    >>> Format.packed.seperated(arg)
    
    will expect the argument *arg* in the packed format and return the same data
    in the seperated format.
    
    Often weights are optional, in this case the class *Format.short* offers
    the two methods *pack* and *unpack*. They return the format listed below
    and expect the parameter in the other format.
    
    - *pack*            [(Feature, Label)]
    - *unpack*          ([Features], [Labels])
    
    
    >>> Format.packed.seperated(([('f1','l1','w1'), ('f2','l2','w2'), ('f3','l3','w3')]))
    (['f1', 'f2', 'f3'], ['l1', 'l2', 'l3'], ['w1', 'w2', 'w3'])
    
    """
    
    @staticmethod
    def get(lst):
        """Determine the format of the argument and return respective class
        
        Depending on the number of samples, and type of the list content, it may
        happen that the format can't be uniquely determined.
        
        :param lst:         List in unkown format
        :returns:           Instance of Format.<list format>
        :raises:            Exception, if the format cannot be determined
        """
        sig = [Format.packed.match(lst), Format.hipack.match(lst), Format.lopack.match(lst), Format.seperated.match(lst)]
        if sum(sig) != 1: raise Exception('Format cannot be determined')
        idx = sig.index(True)
        flst = (Format.packed, Format.hipack, Format.lopack, Format.seperated)
        return flst[idx]()


    
    @staticmethod
    def _pack(head, tail):
        """
        :param head:        [(1,..,n-1)]
        :param tail:        [n]
        :returns:           [(1,..n)]
        """
        if len(head) != len(tail): raise Exception('head and tail must have equal size')
        return [ tuple( list(head[i]) + [tail[i]] ) for i in range(len(head))]
    
    @staticmethod
    def _unpack(lst):
        """
        :param lst:         [(1,..,n-1,n)]
        :returns:           [(1,..,n-1)], [n]
        """
        return reduce(lambda (head,tail), itm: (head + [itm[:-1]], tail + [itm[-1]]), lst, ([],[]))

    class short:
        @staticmethod
        def pack(features, labels):
            """
            :param:         [features], [labels]
            :returns:       [(feature, label)]
            """
            head = [() for i in range(len(features))]
            head = Format._pack(head, features)
            head = Format._pack(head, labels)
            return head
        
        @staticmethod
        def unpack(lst):
            """
            :param:         [(feature, label)]
            :returns:       [features], [labels]
            """
            head, labels = Format._unpack(lst)
            head, features = Format._unpack(head)
            return features, labels

    class packed:
        """[(feature, label, weight)]"""
        @staticmethod
        def packed(lst):
            """
            :param:         [(feature, label, weight)]
            :returns:       [(feature, label, weight)]
            """
            return lst
        @staticmethod
        def hipack(lst):
            """
            :param:         [(feature, label, weight)]
            :returns:       ([(feature, label)], [weights])
            """
            return Format._unpack(lst)
        @staticmethod
        def lopack(lst):
            """
            :param:         [(feature, label, weight)]
            :returns:       ([feature], [(label, weight)])
            """
            return Format.seperated.lopack(Format.packed.seperated(lst))
        @staticmethod
        def seperated(lst):
            """
            :param:         [(feature, label, weight)]
            :returns:       ([features], [labels], [weights])
            """
            head, weights = Format._unpack(lst)
            head, labels = Format._unpack(head)
            features = _tail(Format._unpack(head))
            return features, labels, weights
        @staticmethod
        def match(lst):
            """Check, if the argument *lst* could be of type Format.packed."""
            return isinstance(lst, list) and isinstance(lst[0], tuple) and len(lst[0]) == 3

        
    class hipack:
        """[(feature, label, weight)]"""
        @staticmethod
        def packed((t,weights)):
            """
            :param:         ([(feature, label)], [weights])
            :returns:       [(feature, label, weight)]
            """
            return Format._pack(t, weights)
        @staticmethod
        def hipack((t,weights)):
            """
            :param:         ([(feature, label)], [weights])
            :returns:       ([(feature, label)], [weights])
            """
            return t,weights
        @staticmethod
        def lopack((t,weights)):
            """
            :param:         ([(feature, label)], [weights])
            :returns:       ([feature], [(label, weight)])
            """
            return Format.seperated.lopack(Format.hipack.seperated((t,weights)))
        @staticmethod
        def seperated((t,weights)):
            """
            :param:         ([(feature, label)], [weights])
            :returns:       ([features], [labels], [weights])
            """
            head, labels = Format._unpack(t)
            features = _tail(Format._unpack(head))
            return features, labels, weights
        @staticmethod
        def match(lst):
            """Check, if the argument *lst* could be of type Format.hipack."""
            return isinstance(lst, tuple) and isinstance(lst[0], list) and isinstance(lst[0][0], tuple) and isinstance(lst[1], list) and len(lst) == 2 and len(lst[0][0]) == 2

    class lopack:
        """[(feature, label, weight)]"""
        
        @staticmethod
        def packed((features, lw)):
            """
            :param:         ([feature], [(label, weight)])
            :returns:       [(feature, label, weight)]
            """
            return Format.seperated.packed(Format.lopack.seperated((features, lw)))
        @staticmethod
        def hipack((features, lw)):
            """
            :param:         ([feature], [(label, weight)])
            :returns:       ([(feature, label)], [weights])
            """
            return Format.seperated.hipack(Format.lopack.seperated((features, lw)))
        @staticmethod
        def lopack((features, lw)):
            """
            :param:         ([feature], [(label, weight)])
            :returns:       ([feature], [(label, weight)])
            """
            return features, lw
            
        @staticmethod
        def seperated((features, lw)):
            """
            :param:         ([feature], [(label, weight)])
            :returns:       ([features], [labels], [weights])
            """
            head, weights = Format._unpack(lw)
            labels = _tail(Format._unpack(head))
            return features, labels, weights
            
        @staticmethod
        def match(lst):
            """Check, if the argument *lst* could be of type Format.lopack."""
            return isinstance(lst, tuple) and isinstance(lst[0], list) and isinstance(lst[1], list) and isinstance(lst[1][0], tuple) and len(lst) == 2 and len(lst[1][0]) == 2
    
    class seperated:
        """[(feature, label, weight)]"""
        @staticmethod
        def packed((features, labels, weights)):
            """
            :param:         ([features], [labels], [weights])
            :returns:       [(feature, label, weight)]
            """
            head = [() for i in range(len(features))]
            head = Format._pack(head, features)
            head = Format._pack(head, labels)
            head = Format._pack(head, weights)
            return head
        @staticmethod
        def hipack((features, labels, weights)):
            """
            :param:         ([features], [labels], [weights])
            :returns:       ([(feature, label)], [weights])
            """
            head = [() for i in range(len(features))]
            head = Format._pack(head, features)
            head = Format._pack(head, labels)
            return head, weights
        @staticmethod
        def lopack((features, labels, weights)):
            """
            :param:         ([features], [labels], [weights])
            :returns:       ([feature], [(label, weight)])
            """
            head = [() for i in range(len(features))]
            head = Format._pack(head, labels)
            head = Format._pack(head, weights)
            return features, head
            
        @staticmethod
        def seperated((features, labels, weights)):
            """
            :param:         ([features], [labels], [weights])
            :returns:       ([features], [labels], [weights])
            """
            return features, labels, weights
        
        @staticmethod
        def match(lst):
            """Check, if the argument *lst* could be of type Format.seperated."""
            return isinstance(lst, tuple) and isinstance(lst[0], list) and len(lst) == 3

"""
#Toy data
v1 = [('f1','l1','w1'), ('f2','l2','w2'), ('f3','l3','w3')]     # [(feature, label, weight)]
v2 = [('f1','l1'), ('f2','l2'), ('f3','l3')], ['w1','w2','w3']  # [(feature, label)], [weight]
v3 = ['f1','f2','f3'], [('l1','w1'), ('l2','w2'), ('l3','w3')]  # [features], [(label, weight)]
v4 = ['f1','f2','f3'], ['l1','l2','l3'], ['w1','w2','w3']       # [features], [labels], [weights]
"""

