
import math
import scipy.spatial

_fst = lambda l: l[0]
_snd = lambda l: l[1]

def LTW(x, y, dist=scipy.spatial.distance.euclidean):
    """Linear time warping.
    
    :param x:       First input vector.
    :type x:        [a]
    :param y:       Second input vector.
    :type y:        [b]
    :param dist:    Distance measurement for vector elements.
    :type dist:     (Num c) => a -> b -> c
    :returns:       Linear time warping distance of *x*, *y*.
    
    >>> LTW([1,2,3,4,5,6,7], [9,8,7,6])
    28.0
    
    """
    I = len(x)
    J = len(y)
    w = lambda i: int( (J-1)/(I-1) * (i-1) + 1 + 0.5 )
    d = sum( [dist(x[i], y[w(i)]) for i in range(I)] )
    return d


def DTW(x, y, dist=scipy.spatial.distance.euclidean, w=3):
    """Dynamic time warping.
    
    .. warning:: Does not work
    
    :param x:       First vector.
    :type x:        [a]
    :param y:       Second vector.
    :type y:        [b]
    :param dist:    Distance for vector elements.
    :type dist:     (Num c) => a -> b -> c
    :param w:       Window size.
    :type w:        int. default is abs(len(x) - len(y))
    :returns:       DTW distance of input vectors as float.
    
    """
    
    I = len(x)
    J = len(y)
    inf = float('inf')
    
    w = max( w, abs(I-J) )
    
    A = scipy.matrix([[inf] * J for j in range(I)])
    A[0,0] = dist(x[0], y[0])
    
    for i in range(1,I):
        for j in range(max(1, i-w), min(J, i+w)):
            m = (
                A[i-1, j  ],    # insertion
                A[  i, j-1],    # deletion
                A[i-1, j-1]     # match
            )
            A[i,j] = dist(x[i],y[j]) + min(m)
    return A[I-1,J-1]


def DTW_from_path(x, y, path, dist=scipy.spatial.distance.euclidean):
    """
    
    .. warning:: does not work
    
    """
    return sum([dist(x[i], y[j]) for (i,j) in path])

def DTW_path(x, y, dist=scipy.spatial.distance.euclidean, w=3):
    """Dynamic time warping.
    
    .. warning:: Does not work
    
    :param x:       First vector.
    :type x:        [a]
    :param y:       Second vector.
    :type y:        [b]
    :param dist:    Distance for vector elements.
    :type dist:     (Num c) => a -> b -> c
    :param w:       Window size.
    :type w:        int. default is abs(len(x) - len(y))
    :returns:       DTW distance of input vectors as float.
    
    """
    
    I = len(x)
    J = len(y)
    inf = float('inf')
    
    w = max( w, abs(I-J) )
    
    A = scipy.matrix([[inf] * J for j in range(I)])
    B = scipy.matrix([[inf] * J for j in range(I)])
    A[0,0] = dist(x[0], y[0])
    B[0,0] = 0
    
    for i in range(1,I):
        for j in range(max(1, i-w), min(J, i+w)):
            m = (
                (A[i-1, j  ], j),    # insertion
                (A[  i, j-1], j-1),    # deletion
                (A[i-1, j-1], j-1)     # match
            )
            A[i,j] = dist(x[i],y[j]) + _fst(min(m, key=fst))
            B[i,j] = _snd(min(m, key=fst))
    
    i,j = I-1, J-1
    path = [(i,j)]
    while i > 0 and j > 0:
        c = int(B[i,j])
        if math.isinf(c):
            raise Exception('DTW: Backtracking sequence not legal.')
        i = i - 1
        j = c
        path.append( (i,c) )
    
    #return reversed(path), A[I-1,J-1]
    return path, A[I-1,J-1]


def KLD(p, q, r, save=True):
    """Return the continous Kullback-Leibler divergence.
    
    *q*(x) must be non-zero for all x in *r*. If this doesn't hold for *q*,
    then *inf* is returned. If the *save* flag is set, the support *r* is
    reduced to the support of *q* (thus *q* non-zero for all x in *r'*).
    
    If you have lists instead of functions, use :func:`listFunc` and :func:`range`.
    
    :param p:           pdf of the first distribution.
    :type p:            (Num b) :: a -> b
    :param q:           pdf of the second distribution.
    :type q:            (Num b) :: a -> b
    :param r:           Support.
    :type r:            [a]
    :param save:        Flag, if support may be reduced.
    :type save:         bool
    :returns:           KLD of *p* and *q*, measured at locations in *r*.
    
    >>> KLD(lambda x: x, lambda x: x-1, [0,1,2,3,4], save=True)
    5.4150374992788439
    
    >>> KLD(lambda x: x, lambda x: x-1, [0,1,2,3,4], save=True)
    inf
    
    
    """
    s = 0.0
    for x in r:
        if p(x) != 0.0:
            if q(x) != 0.0:
                s += p(x) * math.log(float(p(x)) / q(x), 2)
            elif q(x) == 0.0 and not save:
                s = float('inf')
                break
    return s

# Levenshtein distance (edit distance)

