
import model
import numpy
import scipy
import scipy.linalg

class LinearModel(model.Model):
    """Generalized linear model base class.
    
    Although possible, it's generally not a good idea to use this class
    directly. Instead, use a derived classes, such as :class:`BasisRegression`.
    
    Subclasses are required to write the *coeff* member after fitting. Otherwise
    some routine calls may fail or produce wrong results.
    
    """
    def __init__(self, lambda_=0.0):
        """
        :param lambda_:     Shrinkage factor.
        :type lambda_:      float
        """
        self.lambda_ = lambda_
    
    def _computeCoeff(self, X, Y, lambda_=0.0, weights=None):
        """Solve least squares linear equation for the coefficients.
        
        b = (X.T * W * X + lambda_ * I)^(-1) * X * W * Y
        
        .. todo::
            If rows and columns are dependent, drop one
            Thus, also one of the coefficients has to be set to zero.
        
        :param X:           Basis inputs
        :type X:            scipy.matrix [Nx(p+1)]
        :param Y:           Output values
        :type Y:            scipy.matrix [NxK]
        :param lambda_:     Shrinkage factor
        :type lambda_:      float
        :param weights:     Weights of samples.
        :type weights:      [float]
        :returns:           (X.T * W * X + lambda * I)^(-1) * X.T * W * Y
        """
        N, M = X.shape
        # N: Number of data points, (p+1): Number of coeffs, K: dimension of output
        # (X.T * X + lambda * I)^-1 * X.T * Y
        # => a = (X.T * X + lambda* I)^-1
        # => b = X.T * Y
        # => a^-1 * b = x => a * a ^-1 * b = a * x => a * x = b (a,b known)
        # ((p+1)xN * Nx(p+1) + I_(p+1))^-1 * ((p+1)xN) * (NxK) = (p+1)xK
        if weights is None: weights = [1.0/N for i in range(N)]
        Id = scipy.matrix(numpy.diag([1.0 for i in range(M)]))
        W = scipy.matrix(numpy.diag(weights))
        a = X.T * W * X + lambda_ * Id
        b = X.T * W * Y
        return scipy.matrix(scipy.linalg.solve(a,b, overwrite_a=True, overwrite_b=True))
    
    def fit(self, data, labels, weights=None):
        """Fit the model to some training data.
        
        :param data:        Training inputs.
        :type data:         [a]
        :param labels:      Training labels.
        :type labels:       [b]
        :param weights:     Sample weights.
        :type weights:      [float]
        :returns:           self
        
        """
        # ([[num]], [num]) => X: Nxp, Y: Nx1
        X = scipy.matrix(map(lambda p: [1.0] + p, data))
        Y = scipy.matrix(labels)
        if Y.shape[0] == 1: Y = Y.T
        self.coeff = self._computeCoeff(X, Y, self.lambda_, weights)
        return self
    
    def eval(self, x):
        """Evaluate the model at data point *x*.
        
        The last computed coefficients are used.
        
        .. todo::
            This routine doesn't work, if *x* is a scipy.matrix
        
        :param x:           (p+1) Basis inputs.
        :type x:            [float]
        :returns:           Model output.
        """
        # 1x(p+1) * (p+1)xK = 1xK => [1.0 + x] * coeff
        # Kx(p+1) * (p+1)x1 = Kx1 => coeff.T * [1.0 + x].T
        y = scipy.matrix([1.0] + x) * self.coeff
        return y
    
    def numOutput(self):
        """Number of output values."""
        return self.coeff.shape[1]
    
    def numCoeff(self):
        """Number of basis values."""
        return self.coeff.shape[0]
    
    def err(self, x, y):
        """Squared residual."""
        e = self.eval(x)
        return (y-e)**2

    def __str__(self):
        """Converts the linear function to its string representation."""
        y = []
        for j in range(self.coeff.shape[1]):
            c = [str(self.coeff[i,j]) + ' * b' + str(i) for i in range(self.coeff.shape[0])]
            y.append( 'y' + str(j) + ' = ' + ' + '.join(c) )
        return 'LinearModel(' + '; '.join(y) + ')'
        


class BasisRegression(LinearModel):
    """Least squares regression with arbitrary basis input.
    
    This class allows to use arbitrary basis functions. The resulting model is
    
     f(x) = c0*b0(x) + c1*b1(x) + c2*b2(x) + c3*b3(x) + ...
    
    with b0(x) := 1.0
    
    The basis functions can be registered using the *addBasis* method. Any
    function can be used, the only requirement is that the type of its only
    argument is the same type as the model input *x* (see below).
    
    """
    def __init__(self, lambda_=0.0):
        """
        :param lambda_:     Shrinkage factor.
        :type lambda_:      float
        """
        LinearModel.__init__(self, lambda_)
        self.basis = []
    
    def addBasis(self, b):
        """Add a basis function.
        
        :param b:           Basis function.
        :type b:            (Num b) :: a -> b
        """
        self.basis.append( b )
    
    def fit(self, data, labels, weights=None):
        """Fit the model to some training data.
        
        :param data:        Training inputs.
        :type data:         a
        :param labels:      Training labels.
        :type labels:       b
        :param weights:     Sample weights.
        :type weights:      [float]
        :returns:           self
        
        """
        Y = scipy.matrix(labels)
        if Y.shape[0] == 1: Y = Y.T
        # apply basis function h_i(X) = X^i
        X = scipy.matrix([[1.0] + [b(x) for b in self.basis] for x in data])
        self.coeff = self._computeCoeff(X, Y, self.lambda_, weights)
        return self
    
    def eval(self, x):
        """Evaluate the model at data point *x*."""
        return LinearModel.eval(self, [b(x) for b in self.basis])[0,0]

class PolynomialRegression(BasisRegression):
    """Polynomial regression.
    
    For fitting a polynomial f(x) = c0 + c1*x + c2*x**2 + c3*x**3 + ...
    
    """
    def __init__(self, deg, lambda_=0.0):
        """
        :param deg:         Degree of polynomial (largest exponent)
        :type deg:          int
        :param lambda_:     Shrinkage factor.
        :type lambda_:      float
        """
        BasisRegression.__init__(self, lambda_)
        for i in range(1, deg+1):
            self.addBasis( lambda x,i=i: x**i ) # i=i required to trick out closure

    def __str__(self):
        """Converts the polynom to its string representation."""
        c = [str(self.coeff[i,0]) + ' * x**' + str(i) for i in range(self.coeff.shape[0])]
        return 'Poly(' + ' + '.join(c) + ')'


"""
TODO
    > Least angle regression
    > Partial least squares
    lasso regression
    Principal component regression
    TODO: Be specific about the input format!
    TODO: If rows/columns are dependent, drop one
    TODO: F-Statistic = ((RSS_0 - RSS_1) / (p_1 - p_0)) / (RSS_1 / (N-p_1-1)), RSS_0,p_0: p_0+1 params, RSS_1,p_1: p_1+1 params, p_1 > p_0

Polynomial alternative implementation:
    class PolynomialRegression(LinearModel):
        def __init__(self, deg, lambda_=0.0):
            LinearModel.__init__(self, lambda_)
            self.degree = deg

        def fit(self, data, labels, weights=None):
            Y = scipy.matrix(labels)
            if Y.shape[0] == 1: Y = Y.T
            # apply basis function h_i(X) = X^i
            X = scipy.matrix([[x**i for i in range(self.degree+1)] for x in data])
            self.coeff = self._computeCoeff(X, Y, self.lambda_, weights)
            return self

        def eval(self, x):
            return LinearModel.eval(self, [x**i for i in range(1,self.degree+1)])[0,0]

Notes
    input X = (X1,X2,..,Xp)
    f(X) = beta_0 + sum( X_j * beta_j, j=1..p )

    Lasso:
     > Convex optimization (cvxopt)

    inference: beta ~ N(beta, (X.T * X)^(-1) * sigma**2)
     With additive gaussian zero mean noise

    z-score
     z_j = beta_j / (sigma * sqrt( v_j ))
     v_j: j'th diagonal element of (X.T*X)^(-1)

    F = ((RSS_0 - RSS_1) / (p_1 - p_0)) / (RSS_1 / (N-p1-1))
     RSS_1: RSS for model with p1 + 1 params
     RSS_0: RSS for model with p0 + 1 params
     p1 > p0: Number of parameters
     => p1 - p0 params constrained to be zero
     => Measures change in RSS per additional parameter
"""


