
from ailib import majorityVote, mean #, histogram
import random

class Model:
    """Interface for fitting algorithm."""
    
    def fit(self, features, labels=None, weights=None):
        """Fit the model to the training data.
        
        :param features:    Training samples.
        :param labels:      Target values.
        :param weights:     Sample weights.
        :type weights:      [float]
        :returns:           self
        """
        abstract()
    
    def eval(self, x):
        """Evaluate the model at data point *x*."""
        abstract()
    
    def err(self, (x,y)):
        """Return the distance between the target *y* and the model
        prediction at the point *x*."""
        abstract()

    def __call__(self, x):
        return self.eval(x)


class Committee(Model):
    """Interface for fitting algorithms that base on an ensemble of several models.
    
    Typically, the committee consists of several --- possibly different ---
    submodels. All submodels are trained first, where the manner in which
    this happens is not known here (thus the *fit* method is not overwritten).
    For prediction, each trained submodel is evaluated. The final result
    is then based on the individual predictions. How the predictions are
    assembled depends on the problem type.
    
    The mixin classes provide common evaluation and error functions.
    Use like so:
    
    >>> class foo(Committee.Sampling, Committee): pass
    
    """
    
    def __init__(self):
        self.models = []
    
    def addModel(self, m):
        """Add a model *m* to the committee."""
        self.models.append( m )


    ### MIXINS ###
    
    class Regression:
        """Mixin for regression models."""
        
        def eval(self, x):
            """Return the average model prediction at *x*."""
            y = [t.eval(x) for t in self.models]
            return mean(y)

        def err(self, (x,y)):
            """Squared residual."""
            e = self.eval(x)
            return (e-y)**2

    class Classification:
        """Mixin for classification models."""
        
        def eval(self, x):
            """Return the most frequently predicted class of *x*."""
            y = [t.eval(x) for t in self.models]
            return majorityVote(y)

        def err(self, (x,y)):
            """:term:`0-1 loss function` for classification."""
            e = self.eval(x)
            if e == y: return 0.0
            return 1.0

    class Sampling:
        """Mixin for classification models. The class label is sampled
        from all predicted labels."""
        
        # Sample from eval distribution
        def eval(self, x):
            """Return the class of *x*, sampled from the predictions
            of the individual models."""
            y = [t.eval(x) for t in self.models]
            return random.choice(y)

        def err(self, (x,y)):
            """:term:`0-1 loss function` for classification."""
            e = self.eval(x)
            if e == y: return 0.0
            return 1.0

