from ailib import Format
import itertools
import model
from numpy import diag
from numpy.linalg import eigh
from scipy import matrix
from scipy.linalg import norm, solve
import sys

class IterLimit(Exception):
    """Is thrown, if the number of iterations exceeds a limit."""
    pass

def posDef(m, strict=True):
    """Check, if a matrix *m* is positive definite.
    
    :param m:           Matrix to be checked.
    :type m:            scipy.matrix
    :returns:           True, if *m* is positive definite.
    
    """
    w,v = eigh(m)
    k = min(w)
    return k > 0 or (not strict and k == 0)

def posSemiDef(m):
    """Check, if a matrix *m* is positive semi-definite.
    
    :param m:           Matrix to be checked.
    :type m:            scipy.matrix
    :returns:           True, if *m* is positive semi-definite.
    
    """
    posDef(m, False)


class NLSQ(model.Model):
    """Non-linear least-squares base class.
    
    This is an incomplete class for nonlinear least-squares algorithms.
    The class provides some basic routines that are commonly used.
    
    Subclasses are required to write the *coeff* member after fitting. Otherwise
    some routine calls may fail or produce wrong results. The default strategy
    for the step size is a constant.
    
    .. warning:: Local minima!
    
    Throughout the documentation of this class, the following type
    symbols are used:
    - a: The parameters of the model. ????????
    - b: The input of the model. May be arbitrary.
    - c: The output of the model. ?????????
    - N: Number of training samples.
    - M: Number of model function parameters.
    
    The search line parameters are:
    - *rho*         ?
    - *beta*        ?
    - *al_max*      ?
    - *tau*         ?
    - *eps*         ?
    
    """
    def __init__(self, model, derivs, initial=None, lsParams=None, sndDerivs=None):
        """
        
        .. warning:: parameter dependence!
        
        :param model:       Model function to be fitted.
        :type model:        (Num c) :: a -> b -> c
        :param derivs:      First order derivations of the model function.
                            Expected is list where the i'th element represents
                            the derivation w.r.t the i'th parameter.
        :type derivs:       [(Num c) :: a -> b -> c]
        :param initial:     Initial choice of parameters.
                            The initial choice may influence result and convergence speed.
        :type initial:      c
        :param lsParams:    Line search parameters
        :type lsParams:     [float]
        :param sndDerivs:   Second order partial derivations of the model
                            function. Expected is a matrix of functions where
                            the element (ij) represents the derivations w.r.t
                            the i'th first and then the j'th parameter. Only
                            required for the :term:`Hessian matrix`.
        :type sndDerivs:    [[(Num c) :: a -> b -> c]]
        
        """
        self.model = model
        self.partial = derivs
        self.numCoeff = len(self.partial)
        if initial is None: initial = matrix([0.0] * self.numCoeff).T
        self.initial = initial
        self.coeff = initial.copy()
        if lsParams is None: lsParams = [0.0001, 0.999, 0.001, 0.001, 0.0000001] # rho in (0,0.25), beta in (rho,1.0), al_max, tau > 0
        self.lsParams = lsParams[:]
        self.sndPartial = sndDerivs

        self.alpha = lambda a,b,c,d: 0.001
        #self.alpha = self._softLineSearch
        #self.alpha = self._exactLineSearch

    def grad(self, theta, data):
        """Compute the :term:`Gradient` of the error function.
        
        The error function is dependent on the model function parameters and
        the training data. This information has to be provided.
        
        .. todo::
            Cannot handle weights.
            Implemented but commented out (not tested, not quite sure)
        
        :param theta:       The parameters of the model function.
        :type theta:        a
        :param data:        The training samples.
        :type data:         :term:`STD`
        :returns:           Gradient as a (Nx1) matrix
        
        """
        g = []
        for j in range(self.numCoeff):
            v = sum([ (y - self.model(theta,x)) * -self.partial[j](theta,x) for (x,y,w) in data ]) # sum([ w * (...
            g.append(v)
        return matrix(g).T

    def jacobian(self, theta, data):
        """Compute the :term:`Jacobian matrix` of the error function.
        
        The error function is dependent on the model function parameters and
        the training data. This information has to be provided.
        
        .. todo::
            Cannot handle weights.
            Actually, as J_ij = [df<j>/dt<i>], weights don't appear
            In the context of Gauss-Newton and Levenberg-Marquandt, J*W is used.
        
        :param theta:       The parameters of the model function
        :type theta:        a
        :param data:        The training samples.
        :type data:         :term:`STD`
        :returns:           Jacobian matrix (NxM)
        
        """
        m = matrix([[ -self.partial[j](theta, x) for j in range(self.numCoeff) ] for (x,y,w) in data ])
        #if not weights is None:
        #    return m * matrix(diag(weights))
        return m
    
    def hessian(self, theta, data):
        """Compute the :term:`Hessian matrix` of the error function.
        
        In order to compute the hessian, the first and second order partial
        derivatives have to be specified (use the constructor's sndDerivs
        parameter).
        The error function is dependent on the model function parameters and
        the training data. This information has to be provided.
        
        .. todo::
            Cannot handle weights.
            Implemented but commented out (not tested, not quite sure)
        
        :param theta:       The parameters of the model function.
        :type theta:        a
        :param data:        The training samples.
        :type data:         :term:`STD`
        :returns:           Hessian matrix (MxM)
        """
        H = []
        for i in range(self.numCoeff):
            h = []
            for j in range(self.numCoeff):
                # generate value
                """
                d^2 F(t) / (d(theta_i) d(theta_j))
                = sum( dM/dtj * dM/dti - (y_i - model(theta, x_i)) * dM/(dti,dtj) )
                = sum( self.partials[j](theta, x) * self.partials[i](theta, x) - (y - self.model(theta, x)) * self.hessPartials[i][j](theta, x) )
                """
                s = 0.0
                #w = itertools.chain(weights)
                for (x,y,w) in data:
                    pi = self.partial[i](theta, x)
                    pj = self.partial[j](theta, x)
                    pij = self.sndPartial[i][j](theta, x)
                    s += (pi * pj - (y  - self.model(theta, x)) * pij) # * w
                h.append(s)
            H.append(h)
        return matrix(H)
    

    def _softLineSearch(self, alpha, theta, dir, data):
        """Compute the step size.
        
        Compute an approximation of the optimal step size, given the error
        function and a moving direction. The approximation is relatively
        lax, as the value of alpha is only modified, if it exceeds some
        boundaries.
        
        .. todo:: parameter handling
        
        :param alpha:       Last used alpha.
        :type alpha:        float
        :param theta:       Current estimate of the model function parameters.
        :type theta:        a
        :param dir:         Moving direction.
        :type dir:          a
        :param data:        Training data.
        :type data:         :term:`STD`
        :param rho:         
        :type rho:          float
        :param beta:        
        :type beta:         float
        :param al_max:      
        :type al_max:       float
        :returns:           The next value for the step size alpha.
        
        """
        rho, beta, al_max = self.lsParams[:3]
        
        # initial interval
        a = 0
        b = min(1, al_max)
        
        phi = lambda alpha: 0.5 * sum([(y - self.model(theta + alpha * dir, x))**2 for (x,y,w) in data])
        dphi = lambda alpha: (dir.T * self.grad(theta + alpha * dir, data))[0,0]
        dpz, pz = dphi(0.0), phi(0.0)
        lambda_ = lambda al: pz + rho * al * dpz
        #if pz >= 0: raise Exception('Predicted direction does not decrease error.')
        
        # enforce lower bound
        while b < al_max and phi(b) <= lambda_(b) and dphi(b) <= beta * dpz:
            print a,b
            a = b
            b = min(2*b, al_max)
        
        # set alpha
        alpha = b
        
        # enforce both constraints
        # infinite loop possible! how & where??
        while phi(alpha) > lambda_(alpha) or dphi(alpha) < beta * dpz:
            alpha, a, b = self._refine( (a,b), phi, dphi, lambda_)
        
        #if phi(alpha) >= pz: raise Exception('Predicted step size does not decrease error.')
        return alpha
    
    def _exactLineSearch(self, alpha, theta, dir, data):
        """Compute the step size.
        
        Compute the optimal step size, given the error function and a moving
        direction. Instead of the very optimal step size, an arbitrary close
        approximation can be computed. Note that the soft line search is often
        preferred over the exact line search, as it's generally faster.
        
        .. warning:: eps parameter used by children.. rename!?
        
        .. todo:: parameter handling
        
        :param alpha:       Last used alpha.
        :type alpha:        float
        :param theta:       Current estimate of the model function parameters.
        :type theta:        a
        :param dir:         Moving direction.
        :type dir:          a
        :param data:        Training data.
        :type data:         :term:`STD`
        :param rho:         
        :type rho:          float
        :param beta:        
        :type beta:         float
        :param al_max:      
        :type al_max:       float
        :param tau:         
        :type tau:          float
        :param eps:         
        :type eps:          float
        :returns:           The next value for the step size alpha.
        
        """
        rho,beta,al_max,tau,eps = self.lsParams[:5]
        
        # initial interval
        a = 0
        b = min(1, al_max)
        
        phi = lambda alpha: 0.5 * sum([(y - self.model(theta + alpha * dir, x))**2 for (x,y,w) in data])
        dphi = lambda alpha: (dir.T * self.grad(theta + alpha * dir, data))[0,0]
        dpz, pz = dphi(0.0), phi(0.0)
        lambda_ = lambda al: pz + rho * al * dpz
        #if pz >= 0: raise Exception('Predicted direction does not decrease error.')
        
        # enforce lower bound
        while phi(b) <= lambda_(b) and dphi(b) <= beta * dpz and b < al_max:
            a = b
            b = min(2*b, al_max)
        
        # set alpha
        alpha = b
        
        # enforce both constraints
        while b - a > eps and abs(dphi(alpha)) > tau * abs(dpz):
            alpha, a, b = self._refine( (a,b), phi, dphi, lambda_)
        
        #if phi(alpha) >= pz: raise Exception('Predicted step size does not decrease error.')
        return alpha

    def _refine(self, (a,b), phi, dphi, lambda_):
        """Narrow the interval *(a,b)* of acceptable values for the step size alpha.
        
        :param (a,b):       Current interval limits.
        :type (a,b):        (float,float)
        :param phi:         
        :type phi:          float -> c
        :param dphi:        
        :type dphi:         float -> c
        :param lambda_:     
        :type lambda_:      float -> c
        :returns:           The adjusted interval *(a,b)* and step size alpha.
        
        """
        D = b - a
        c = (phi(b) - phi(a) - D * dphi(a) ) / D**2
        
        # compute alpha
        if c > 0:
            alpha = a - dphi(a) / (2*c)
            alpha = min(max(alpha, a + 0.1 * D), b - 0.1 * D)
        else:
            alpha = (a+b)/2
        
        # adjust interval
        if phi(alpha) < lambda_(alpha):
            a = alpha
        else:
            b = alpha
        
        return alpha, a, b

    def _gainRatio(self, dir, mu, theta, data):
        """Compute the :ref:`Gain ratio<nlsq-gain-ratio>` of a moving direction.
        
        .. todo:: Doesn't support weights so far.
        
        :param dir:         Moving direction.
        :type dir:          a
        :param mu:          Damping factor.
        :type mu:           float
        :param theta:       Model function parameter estimate.
        :type theta:        a
        :param data:        Training data.
        :type data:         :term:`STD`
        :returns:           The gain ratio of the moving direction.
        """
        F = lambda t: 0.5 * sum([(y - self.model(t,x))**2 for (x,y,w) in data])
        nom = F(theta) - F(theta + dir)
        denom = 0.5 * dir.T * ( mu * dir - self.grad(theta, data) )
        return nom / denom[0,0]

    def eval(self, x):
        """Evaluate the model at data point *x* with coefficients after fitting."""
        return self.model(self.coeff, x)

    def err(self, (x,y)):
        """Squared residual."""
        return (y - self.eval(x))**2


class SteepestDescent(NLSQ):
    """
    Implementation of the :ref:`nlsq-steepest-descent` algorithm.
    
    The initial parameter values and convergence threshold have to be set either
    at object construction or at the call to the *fit* method. The latter
    overwrites the former.
    
    The default strategy for step size computation is the
    :ref:`Soft line search <nlsq-line-search>`. This behaviour can be modified
    through the *alpha* member. Example:
    
    >>> o = SteepestDescent(...)
    >>> o.alpha = o._exactLineSearch
    
    If the fitting fails due to an OverflowError, try modifying the line search
    parameters. Consider setting *al_max* to a relatively small value (<< 1.0),
    although this will increase convergence time.
    """
    def __init__(self, model, derivs, initial=None, eps=0.001, maxIter=100, lsParams=None):
        """
        :param model:       Model function to be fitted.
        :type model:        (Num c) :: a -> b -> c
        :param derivs:      First order derivations of the model function.
                            Expected is list where the i'th element represents
                            the derivation w.r.t the i'th parameter.
        :type derivs:       [(Num c) :: a -> b -> c]
        :param initial:     Initial choice of parameters.
                            The initial choice may influence result and convergence speed.
        :type initial:      c
        :param eps:         Convergence threshold. The iteration stops, if the
                            step length is below this threshold.
        :type eps:          float
        :param maxIter:     Maximum number of iterations.
        :type maxIter:      int
        :param lsParams:    Line search parameters.
        :type lsParams:     [float]
        """
        NLSQ.__init__(self, model, derivs, initial=initial, lsParams=lsParams)
        self.eps = eps
        self.maxIter = maxIter
        self.alpha = self._softLineSearch
    
    def fit(self, samples, labels, weights=None, theta=None, eps=None):
        """Determine the parameters of a model, s.t. it optimally fits the training data.
        
        The parameters *theta* and *eps* are optional. If one of the is not
        set, its default value determined at object construction will be used
        (use the constructor's *initial* and *eps* arguments).
        
        :param samples:     Training data.
        :type samples:      []
        :param labels:      Training labels.
        :type labels:       []
        :param weights:     Sample weights.
        :type weights:      [float]
        :param theta:       Initial parameter choice.
        :type theta:        a
        :param eps:         Convergence threshold. The iteration stops, if the
                            step length is below this threshold.
        :type eps:          float
        :returns:           self
        
        """
        if weights is None: weights = [1.0] * len(samples)
        data = Format.seperated.packed((samples, labels, weights))
        if theta is None: theta = self.initial
        if eps is None: eps = self.eps
        
        numIter = self.maxIter
        alpha = self.lsParams[2]
        while numIter > 0:
            numIter -= 1
            
            h = - self.grad(theta, data)
            alpha = self.alpha(alpha, theta, h, data)
            theta = theta + alpha * h
            if norm(h) <= eps: break # TODO: Break before or after assignment to theta?
            print self.maxIter - numIter, theta.T, alpha
        
        self.coeff = theta
        if numIter <= 0: raise IterLimit()
        return self

class Newton(NLSQ):
    """
    Implementation of :ref:`nlsq-newton`.
    
    Parameter handling and step size strategy behave as in :class:`SteepestDescent`
    """
    def __init__(self, model, derivs, sndDerivs, initial=None, eps=None, maxIter=100, lsParams=None):
        """
        :param model:       Model function to be fitted.
        :type model:        (Num c) :: a -> b -> c
        :param derivs:      First order derivations of the model function.
                            Expected is list where the i'th element represents
                            the derivation w.r.t the i'th parameter.
        :type derivs:       [(Num c) :: a -> b -> c]
        :param sndDerivs:   Second order partial derivations of the model function.
        :type sndDerivs:    [[(Num c) :: a -> b -> c]]
        :param initial:     Initial choice of parameters.
                            The initial choice may influence result and convergence speed.
        :type initial:      c
        :param eps:         Convergence threshold. The iteration stops, if the
                            step length is below this threshold.
        :type eps:          float
        :param maxIter:     Maximum number of iterations.
        :type maxIter:      int
        :param lsParams:    Line search parameters
        :type lsParams:     [float]
        """
        NLSQ.__init__(self, model, derivs, initial=initial, lsParams=lsParams, sndDerivs=sndDerivs)
        if eps is None: eps = 0.01
        self.eps = eps
        self.maxIter = maxIter
        self.alpha = self._softLineSearch
    
    def fit(self, samples, labels, weights=None, theta=None, eps=None):
        """Determine the parameters of a model, s.t. it optimally fits the training data.
        
        The parameters *theta* and *eps* are optional. If one of the is not
        set, its default value determined at object construction will be used
        (use the constructor's *initial* and *eps* arguments).
        
        :param samples:     Training data.
        :type samples:      []
        :param labels:      Training labels.
        :type labels:       []
        :param weights:     Sample weights.
        :type weights:      [float]
        :param theta:       Initial parameter choice.
        :type theta:        a
        :param eps:         Convergence threshold. The iteration stops, if the
                            step length is below this threshold.
        :type eps:          float
        :returns:           self
        
        """
        if weights is None: weights = [1.0] * len(samples)
        data = Format.seperated.packed((samples, labels, weights))
        if theta is None: theta = self.initial
        if eps is None: eps = self.eps
        
        numIter = self.maxIter
        alpha = self.lsParams[2]
        while numIter > 0:
            numIter -= 1
            
            # solve H * h = -J.T * f
            # TODO: Use linearModel?
            a = self.hessian(theta, data)
            b = -self.grad(theta, data)
            h = matrix(solve(a,b, overwrite_a=True, overwrite_b=True))

            # step update
            alpha = self.alpha(alpha, theta, h, data)
            theta = theta + alpha * h
            if norm(h) <= eps: break
            print self.maxIter - numIter, theta.T, alpha
        
        self.coeff = theta
        if numIter <= 0: raise IterLimit()
        return self

class HybridSDN(Newton):
    """
    Implementation of the :ref:`Hybrid SD/Newton <nlsq-newton>`
    algorithm.
    
    The :ref:`Newton <nlsq-newton>` step is chosen, if the :term:`Hessian matrix`
    is positive definite. Otherwise, the
    :ref:`steepest descent method <nlsq-steepest-descent>` is applied.
    
    Parameter handling and step size strategy behave as in :class:`SteepestDescent`
    """
    def __init__(self, model, derivs, sndDerivs, initial=None, eps=None, maxIter=100, lsDescent=None, lsNewton=None):
        """
        :param model:       Model function to be fitted.
        :type model:        (Num c) :: a -> b -> c
        :param derivs:      First order derivations of the model function.
                            Expected is list where the i'th element represents
                            the derivation w.r.t the i'th parameter.
        :type derivs:       [(Num c) :: a -> b -> c]
        :param sndDerivs:   Second order partial derivations of the model function.
        :type sndDerivs:    [[(Num c) :: a -> b -> c]]
        :param initial:     Initial choice of parameters.
                            The initial choice may influence result and convergence speed.
        :type initial:      c
        :param eps:         Convergence threshold. The iteration stops, if the
                            step length is below this threshold.
        :type eps:          float
        :param maxIter:     Maximum number of iterations.
        :type maxIter:      int
        :param lsDescent:   Line search parameters for the steepest descent part.
        :type lsDescent:    [float]
        :param lsNewton:    Line search parameters for the newton part.
        :type lsNewton:     [float]
        """
        Newton.__init__(self, model, derivs, sndDerivs, initial=initial, eps=eps, maxIter=maxIter, lsParams=None)
        if lsNewton is None: lsNewton = self.lsParams
        if lsDescent is None: lsDescent = self.lsParams
        self.lsNewton = lsNewton
        self.lsDescent = lsDescent
    
    def fit(self, samples, labels, weights=None, theta=None, eps=None):
        """Determine the parameters of a model, s.t. it optimally fits the training data.
        
        The parameters *theta* and *eps* are optional. If one of the is not
        set, its default value determined at object construction will be used
        (use the constructor's *initial* and *eps* arguments).
        
        :param samples:     Training data.
        :type samples:      []
        :param labels:      Training labels.
        :type labels:       []
        :param weights:     Sample weights.
        :type weights:      [float]
        :param theta:       Initial parameter choice.
        :type theta:        a
        :param eps:         Convergence threshold. The iteration stops, if the
                            step length is below this threshold.
        :type eps:          float
        :returns:           self
        
        """
        if weights is None: weights = [1.0] * len(samples)
        data = Format.seperated.packed((samples, labels, weights))
        if theta is None: theta = self.initial
        if eps is None: eps = self.eps
        
        numIter = self.maxIter
        alpha = self.lsParams[2]
        while numIter > 0:
            numIter -= 1
            
            H = self.hessian(theta, data)
            g = -self.grad(theta, data)
            if posDef(H):
                sys.stdout.write(str(self.maxIter - numIter) + "  Going newton ")
                # solve H * h = -grad
                h = matrix(solve(H,g, overwrite_a=True, overwrite_b=True))
                self.lsParams = self.lsNewton
            else:
                sys.stdout.write(str(self.maxIter - numIter) + "  Going steepest descent ")
                # h = -grad
                h = g
                self.lsParams = self.lsDescent
            
            alpha = self.alpha(alpha, theta, h, data)
            theta = theta + alpha * h
            print theta.T, alpha
            if norm(h) <= eps: break
        
        self.coeff = theta
        if numIter <= 0: raise IterLimit();
        return self


class GaussNewton(NLSQ):
    """
    Implementation of the :ref:`nlsq-gauss-newton`.
    
    Parameter handling and step size strategy behave as in :class:`SteepestDescent`
    """
    def __init__(self, model, derivs, initial=None, eps=None, maxIter=100, lsParams=None):
        """
        :param model:       Model function to be fitted.
        :type model:        (Num c) :: a -> b -> c
        :param derivs:      First order derivations of the model function.
                            Expected is list where the i'th element represents
                            the derivation w.r.t the i'th parameter.
        :type derivs:       [(Num c) :: a -> b -> c]
        :param initial:     Initial choice of parameters.
                            The initial choice may influence result and convergence speed.
        :type initial:      c
        :param eps:         Convergence threshold. The iteration stops, if the
                            step length is below this threshold.
        :type eps:          float
        :param maxIter:     Maximum number of iterations.
        :type maxIter:      int
        :param lsParams:    Line search parameters
        :type lsParams:     [float]
        """
        NLSQ.__init__(self, model, derivs, initial=initial, lsParams=lsParams)
        if eps is None: eps = 0.01
        self.eps = eps
        self.maxIter = maxIter
        self.alpha = self._softLineSearch

    def fit(self, samples, labels, weights=None, theta=None, eps=None):
        """Determine the parameters of a model, s.t. it optimally fits the training data.
        
        The parameters *theta* and *eps* are optional. If one of the is not
        set, its default value determined at object construction will be used
        (use the constructor's *initial* and *eps* arguments).
        
        :param samples:     Training data.
        :type samples:      []
        :param labels:      Training labels.
        :type labels:       []
        :param weights:     Sample weights.
        :type weights:      [float]
        :param theta:       Initial parameter choice.
        :type theta:        a
        :param eps:         Convergence threshold. The iteration stops, if the
                            step length is below this threshold.
        :type eps:          float
        :returns:           self
        
        """
        if weights is None: weights = [1.0] * len(samples)
        data = Format.seperated.packed((samples, labels, weights))
        if theta is None: theta = self.initial
        if eps is None: eps = self.eps
        
        numIter = self.maxIter
        alpha = self.lsParams[2]
        while numIter > 0:
            numIter -= 1
            
            # solve (J.T * J) * h = -J.T * f
            # check, if J has full rank
            # TODO: Use LinearModel?
            # TODO: Weights
            J = self.jacobian(theta, data)
            a = J.T * J
            b = -self.grad(theta, data)
            h = matrix(solve(a,b, overwrite_a=True, overwrite_b=True))

            # step update
            alpha = self.alpha(alpha, theta, h, data)
            theta = theta + alpha * h
            if norm(h) <= eps: break
            print self.maxIter - numIter, theta.T, alpha
        
        self.coeff = theta
        if numIter <= 0: raise IterLimit()
        return self

class Marquardt(NLSQ):
    """
    Implementation of the :ref:`nlsq-levenberg-marquardt`.
    
    The damping parameter :math:`mu` is initialized with *tau* and adjusted
    according to :ref:`Nielsen's strategy <nlsq-nielsen>`.
    
    """
    def __init__(self, model, derivs, initial=None, eps0=0.01, eps1=0.01, tau=1.0, maxIter=100):
        """
        :param model:       Model function to be fitted.
        :type model:        (Num c) :: a -> b -> c
        :param derivs:      First order derivations of the model function.
                            Expected is list where the i'th element represents
                            the derivation w.r.t the i'th parameter.
        :type derivs:       [(Num c) :: a -> b -> c]
        :param initial:     Initial choice of parameters. The initial choice
                            may influence result and convergence speed.
        :type initial:      c
        :param eps0:        Convergence threshold. The iteration stops, if the
                            maximal gradient component is below this threshold.
        :type eps0:         float
        :param eps1:        Convergence threshold. The iteration stops, if the
                            relative change in the model parameters is below
                            this threshold.
        :type eps1:         float
        :param tau:         Constant for determining the initial damping.
                            Use a small value (10**(-6)), if the initial
                            parameters are believed to be a good approximation,
                            otherwise use 10**(-3) or 1.0 (default)
        :type tau:          float
        :param maxIter:     Maximum number of iterations.
        :type maxIter:      int
        """
        NLSQ.__init__(self, model, derivs, initial=initial)
        self.eps = (eps0, eps1)
        self.maxIter = maxIter
        self.tau = tau
    
    def fit(self, samples, labels, weights=None, theta=None, eps0=None, eps1=None, tau=None):
        """Determine the parameters of a model, s.t. it optimally fits the training data.
        
        The parameters *theta*, *eps0*, *eps1* and *tau* are optional.
        If one of the is not set, its default value determined at object
        construction will be used
        (use the constructor's *initial*, *eps0*, *eps1* and *tau* arguments).
        
        :param samples:     Training data.
        :type samples:      []
        :param labels:      Training labels.
        :type labels:       []
        :param weights:     Sample weights.
        :type weights:      [float]
        :param theta:       Initial parameter choice.
        :type theta:        a
        :param eps0:        Convergence threshold. The iteration stops, if the
                            maximal gradient component is below this threshold.
        :type eps0:         float
        :param eps1:        Convergence threshold. The iteration stops, if the
                            relative change in the model parameters is below
                            this threshold.
        :type eps1:         float
        :param tau:         Constant for determining the initial damping.
                            Use a small value (10**-6), if the initial
                            parameters are believed to be a good approximation,
                            otherwise use 10^-3 or 1.0 (default)
        :type tau:          float
        :returns:           self
        
        """
        if weights is None: weights = [1.0] * len(samples)
        data = Format.seperated.packed((samples, labels, weights))
        if theta is None: theta = self.initial
        if eps0 is None: eps0 = self.eps[0]
        if eps1 is None: eps1 = self.eps[1]
        if tau is None: tau = self.tau
        
        # init
        v = 2
        J = self.jacobian(theta, data)
        J = J.T * J
        mu = tau * max([ J[i,i] for i in range(self.numCoeff) ])
        numIter = self.maxIter
        
        # iterate
        while numIter > 0:
            numIter -= 1
            
            # solve (J.T * J + mu * I) * h = -J * f
            # TODO: Use Linear model?
            # TODO: Speedup?
            # TODO: Weights
            J = self.jacobian(theta, data)
            I = matrix(diag([1.0 for i in range(self.numCoeff)]))
            b = -self.grad(theta, data)
            a = J.T * J + mu * I
            h = matrix(solve(a,b, overwrite_a=True, overwrite_b=True))
            
            # convergence criterion 1: gradient small enough
            if abs(min([b[i,0] for i in range(self.numCoeff)])) <= eps0: break
            
            # convergence criterion 2: step small enough
            if norm(h) <= eps1 * (norm(theta) + eps1): break
            
            # damping adjustment
            r = self._gainRatio(h, mu, theta, data)
            if r > 0:
                mu = mu * max(1.0/3.0, 1 - ( 2*r - 1)**3)
                v = 2
                theta = theta + h
            else:
                mu = mu * v
                v = 2*v
            
            #print self.maxIter - numIter, theta.T, mu
            #sys.stdout.write(str(self.maxIter-numIter) + ' ')
            #sys.stdout.flush()
        
        #print ''
        
        self.coeff = theta
        if numIter <= 0: raise IterLimit()
        return self



"""
Some development help:

model               M
coeff               theta [M]
data                x,y [N]
f(theta, x_i, y_i)  = y_i - M(theta, x_i)
F(theta)            = 0.5 * sum(f(theta, x_i, y_i)**2, i) = 0.5 * sum((y_i - M(theta, x_i))**2, i)
phi(alpha)          = F(theta + alpha * h)
df_i/d(theta_j)     = - dM(theta, x_i) / d(theta_j)
dF/d(theta_j)       = sum( f_i(theta) * -dM(theta,x_i)/d(theta_j), i )
grad(theta)         = [ dF/d(theta_0), dF/d(theta_1), ... ] = nabla(theta)
phi'(alpha)         = h.T * grad( theta + alpha * h)

Provided B (int), M (t -> x -> y), dM/dt<j> (t -> x -> y)
Assumed types: theta (matrix), y (float)

f                   = lambda t, x, y: y - M(t,x)
F                   = lambda t, data: 0.5 * sum([f(t,x,y)**2 for (x,y) in data])
F                   = lambda t, data: 0.5 * sum([(y - M(t,x))**2 for (x,y) in data])
df/dt<j>            = lambda t, x: -dM/dt<j>(t,x)
dF/dt<j>            = lambda t, data: sum([ f(t,x,y) * -dM/dt<j>(t,x) for (x,y) in data])
dF/dt<j>            = lambda t, data: sum([ f(t,x,y) * df/dt<j>(t,x) for (x,y) in data])
grad                = lambda t, data: [dF/dt<j>(t, data) for j in range(B)]
jacobi              = lambda t,data: matrix([df/dt<j>(t,x) for j in range(B) for (x,y) in data])
phi                 = lambda alpha, t, h, data: F(t + alpha * h, data)
dphi                = lambda alpha, t, h, data: h.T * grad(t + alpha * h, data)
"""
