
import model
import copy
import random
import math
from ..ailib import argmin
from ..ailib import mean
from ..ailib import tail, head
from ..ailib import histogram, majorityVote


class Tree(model.Model):
    """Decision tree skeleton.
    
    .. warning:: Don't use this class directly. Use :class:`RegressionTree` or :class:`ClassificationTree` instead.
    
    :param leafThres:   Minimum number of data points per node.
    :type leafThres:    int
    """

    def _computeLabel(self, data):
        """
        """
        abstract()
    def _split(self, data):
        """
        """
        abstract()
    def costSplit(self, j, s):
        """
        """
        abstract()
    def costTree(self, alpha):
        """
        """
        abstract()
    def costTerminal(self, data):
        """
        """
        abstract()


    def __init__(self, leafThres=5):
        self.left  = None
        self.right = None
        self.__data = []
        self.dim = 0
        self.thres = 0
        self.label = 0
        self.leafThres = leafThres
    
    ### Representation ###
    def __str__(self):
        """Convert the tree recursively into a string representation."""
        if self.isLeaf():
            #return "x\t\tl=" + str(self.label) + "\td:" + str(self.data())
            return "x\t\tl=" + str(self.label) + "\td:" + str(len(self.data()))
        else:
            s = "o\t\tj=" + str(self.dim) + "\ts=" + str(self.thres)
            s += "\n|-" + str(self.left).replace('\n','\n  ')
            s += "\n|-" + str(self.right).replace('\n','\n  ')
            return s

    def cleanData(self):
        """Remove the data from the tree.
        
        Only the data member is cleared. The node label is recomputed
        first.
        
        """
        if self.isLeaf():
            self._computeLabel(self.__data)
            self.__data = []
        else:
            self.left.cleanData()
            self.right.cleanData()
    
    ### Basic tree methods ####
    def isLeaf(self):
        """Return true, if the node has no childs (is a leaf node)."""
        return self.left == None and self.right == None
    
    def data(self):
        """Return the data of all child nodes."""
        if self.isLeaf(): return self.__data
        return self.left.data() + self.right.data()
    
    def leaves(self):
        """Return the number of leaves."""
        if self.isLeaf(): return 1
        return self.left.leaves() + self.right.leaves()
    
    def depth(self):
        """Return the maximal depth of the tree."""
        if self.isLeaf(): return 0
        return max(self.left.depth(), self.right.depth()) + 1
    
    def collude(self):
        """Collapses the childs of the node.
        
        The childs' data is collected in this node and the label
        recomputed. The childs will be deleted. The node will become
        a leaf.
        """
        if self.isLeaf(): return
        
        self.__data = self.left.data() + self.right.data()
        self._computeLabel(self.__data)
        del self.left
        del self.right
        self.left = None
        self.right = None
    
    ### Decision tree operations ###
    def prune(self, alpha):
        """Prune the tree to find the tree with minimal cost.
        
        The tree is pruned according to weakest link pruning
        until there's only one node left. Then, the tree with minimal
        total cost is restored.
        
        :param alpha:   Pruning parameter. Controls, how much the tree size influences its cost.
        :type alpha:    float
        """
        tree = copy.deepcopy(self)
        seq = [self.costTree(alpha)]
        while tree.depth() > 0:
            next, c_cost = tree.__minCostCollude()
            next.collude()
            seq.append( tree.costTree(alpha) )
        
        minIdx = min(range(len(seq)), key=lambda i: seq[i])
        for i in range(minIdx):
            head(self.__minCostCollude()).collude()
    
    def __minCostCollude(self):
        """Compute how much a collusion of this node would increase the cost
        of the whole tree.
        
        cost(newtree) =   cost( oldtree )
                        - sum( cost(childs) )
                        + cost( colluded node )
        This method returns the the tree cost difference, i.e.
         cost( colluded node ) - sum( cost(childs) )
        
        """
        cost = - self._terminalSum() + self.costTerminal(self.data())
        
        l = [(self, cost)]
        if not self.left.isLeaf(): l.append( self.left.__minCostCollude() )
        if not self.right.isLeaf(): l.append( self.right.__minCostCollude() )
        return min(l, key=lambda (i,c):c)
    
    def _terminalSum(self):
        """Compute the cost of the tree."""
        if self.isLeaf():
            return self.costTerminal(self.__data)
        else:
            return self.left._terminalSum() + self.right._terminalSum()

    def grow(self, data):
        """Grow a tree on training data.
        
        Finds the optimal splitting parameters of the data and grows
        two child trees (if there are enough data points left).
        
        :param data:    Training data.
        :type data:     :term:`STD`
        """
        if len(data) < self.leafThres:
            self._computeLabel(data)
            self.__data = data
            return
        
        # find j (dim), s (thres)
        D = len(data[0][0])
        splits = []
        for j in range(D):
            # possible splitting points are all x[j] without the rightmost (-> right list would be empty)
            s_pts = map(lambda (x,y): x[j], data)
            s_pts = filter(lambda x: x < max(s_pts),s_pts)
            
            # find the optimal splitting threshold s for the current dimension j
            s_min = argmin(lambda s: self.costSplit(data,j,s), s_pts)
            splits.append( (self.costSplit(data,j,s_min),j,s_min) )
        
        # for all dimensions j, find the best combination of (j,s)
        self.dim, self.thres = tail(min(splits, key=lambda(v,j,s): v))
        self._split(data)

    ### Interface ###
    def eval(self, x):
        """Find the label of a data point *x*."""
        if self.isLeaf(): return self.label
        
        if x[self.dim] <= self.thres:
            return self.left.eval(x)
        else:
            return self.right.eval(x)
    



class Stump(model.Model):
    """
    
    
    
    :param labelValue:  Label of the class.
    :type labelValue:   float
    """

    def __init__(self, labelValue=1.0):
        self.dim = 0
        self.thres = 0.0
        self.labelValue = labelValue

    def fit(self, data, weights=None):
        """Train the stump on the presented data.
        
        Finds the optimal dimension and splitting point to split the data in
        two subsets. The optimal parameters are found w.r.t. to the data
        weights. If no weights are given, a uniform distribution is assumed.
        
        :param data: Data points.
        :type data: :term:`STD`
        :param weights: Weights of data points. Needs not to be normalized.
        :type weights: [float]
        """
        # feature: [num], len(feature) >= self.dim
        
        if weights == None: weights = [1.0/len(data) for i in range(len(data))]
        
        N = len(data)
        D = len(data[0][0])
        
        # for all dimensions j: find optimal thres t, labelValue m
        f = lambda t,j,m: sum([weights[i] for i in filter(lambda i: self.__classify(data[i][0], j, t, m) != data[i][1], range(N))])
        cost = []
        for j in range(D):
            for m in (-self.labelValue, self.labelValue):
                arg = map(lambda d: d[0][j], data)
                t_min = argmin(lambda t: f(t,j,m), arg)
                cost.append( (f(t_min,j,m), j,m,t_min) )

        # select the dimension with the minimal cost
        params = min(cost, key=lambda (s,j,m,t): s)
        self.dim, self.labelValue, self.thres = params[1:]
        return self

    def __classify(self, x, j, t, m):
        """Classify a data point *x* according to the stump parameters
        *j* (dimension), *t* (threshold) and *m* (label).
        """
        if x[j] > t: return float(m)
        return -1.0 * m

    def eval(self, x):
        """Compute the label of a data point *x*."""
        return self.__classify(x, self.dim, self.thres, self.labelValue)

    def err(self, (x,y)):
        """:term:`0-1 loss function`."""
        e = self.eval(x)
        if e == y: return 0.0
        return 1.0

    def __str__(self):
        return "Stump(x[" + str(self.dim) + "] > " + str(self.thres) + " => " + str(self.labelValue) + ")"

class RegressionTree(Tree):
    """

    :param alpha:
    :type alpha: float
    :param leafThres:
    :type leafThres: int
    
    """

    def __init__(self, alpha=0.0, leafThres=5):
        Tree.__init__(self, leafThres)
        self.alpha = alpha
    
    ### Abstract overloads ####

    def _split(self, data):
        """
        """
        self.left = RegressionTree(self.alpha, self.leafThres)
        self.right = RegressionTree(self.alpha, self.leafThres)
        self.left.grow( filter(lambda (x,l): x[self.dim] <= self.thres, data) )
        self.right.grow( filter(lambda (x,l): x[self.dim] > self.thres, data) )
    
    def _computeLabel(self, data):
        """
        """
        self.label = mean(map(lambda (x,y): y, data))
    
    # inner function sum( y[i] - c<i> for i in Rm ) for i=(1,2)
    def costSplit(self, data, j, s):
        """
        """
        #ly = map(lambda (x,y): y, filter(lambda (x,y): x[j] <= s, data))
        #ry = map(lambda (x,y): y, filter(lambda (x,y): x[j] > s, data))
        #c1 = mean(ly) # future label of left
        #c2 = mean(ry) # future label of right
        # costTerminal on future nodes
        #return sum(map(lambda y: (y-c1)**2, ly)) + sum(map(lambda y: (y-c2)**2, ry))
        
        ly = filter(lambda (x,y): x[j] <= s, data)
        ry = filter(lambda (x,y): x[j] > s, data)
        return self.costTerminal(ly) + self.costTerminal(ry)
    
    def costTree(self, alpha):
        """
        """
        return self._terminalSum() + alpha * self.leaves()
    
    def costTerminal(self, data):
        """
        """
        c = mean(map(lambda (x,y): y, data))
        q = mean(map(lambda (x,y): (y-c)**2, data)) # impurity measure
        return len(data)*q

    ### Interface ###
    def fit(self, data):
        """
        """
        # data :: [([num],num)]; [(feature, label)]
        self.grow(data)
        self.prune(self.alpha)
        return self
    
    def err(self, (x,y)):
        """
        """
        e = self.eval(x)
        return (e-y)**2
    
    

class ClassificationTree(Tree):
    """

    :param alpha:
    :type alpha: float
    :param leafThres:
    :type leafThres: int
    """

    def __init__(self, alpha=0.0, leafThres=5):
        Tree.__init__(self, leafThres)
        self.alpha = alpha
    
    ### Abstract overloads ####

    def _split(self, data):
        """
        """
        self.left = ClassificationTree(self.alpha, self.leafThres)
        self.right = ClassificationTree(self.alpha, self.leafThres)
        self.left.grow( filter(lambda (x,l): x[self.dim] <= self.thres, data) )
        self.right.grow( filter(lambda (x,l): x[self.dim] > self.thres, data) )
    
    def _computeLabel(self, data):
        """
        """
        self.label = majorityVote(map(lambda (x,y): y, data))
        # alternative: sample from label distribution
    
    # inner function sum( y[i] - c<i> for i in Rm ) for i=(1,2)
    def costSplit(self, data, j, s):
        """
        """
        ly = filter(lambda (x,y): x[j] <= s, data)
        ry = filter(lambda (x,y): x[j] > s, data)
        return self.costTerminal(ly) + self.costTerminal(ry)
    
    def costTree(self, alpha):
        """
        """
        return self._terminalSum() + alpha * self.leaves()
    
    def costTerminal(self, data):
        """
        """
        hist = histogram(map(lambda (x,y):y, data))
        p = [1.0/len(data) * hist[k] for k in hist]

        # misclassification
        #return 1.0 - 1.0/len(data) * hist[self.label]

        # gini
        return sum([ w * (1 - w) for w in p ])

        # deviance
        #return -sum([ w * log(w) for w in p ])

    ### Interface ###
    def fit(self, data):
        """
        """
        # data :: [([num],num)]; [(feature, label)]
        self.grow(data)
        self.prune(self.alpha)
        return self
    
    def err(self, (x,y)):
        """
        """
        e = self.eval(x)
        if e == y: return 0.0
        return 1.0


# in grow: use samples in main loop
class ForestTreeMixin:
    """
    """
    
    class Regression:
        """
        """
        
        def numDim(self, D):
            """
            """
            # regression; min node size = 5
            return int(math.floor(D/3.0))
    
    class Classification:
        """
        """
        
        def numDim(self, D):
            """
            """
            # classification; min. node size = 1
            return int(math.floor(math.sqrt(D)))
    
    def grow(self, data):
        """
        """
        if len(data) < self.leafThres:
            self._computeLabel(data)
            self.__data = data
            return

        # select m variables at random from D data vars
        D = len(data[0][0])
        m = self.numDim(D)
        sample = random.sample( range(D), m )
        
        # find j (dim), s (thres)
        splits = []
        for j in sample:
            # possible splitting points are all x[j] without the rightmost (-> right list would be empty)
            s_pts = map(lambda (x,y): x[j], data)
            s_pts = filter(lambda x: x < max(s_pts),s_pts)
            s_min = argmin(lambda s: self.costSplit(data,j,s), s_pts)
            splits.append( (self.costSplit(data,j,s_min),j,s_min) )
        
        self.dim, self.thres = tail(min(splits, key=lambda(v,j,s): v))
        self._split(data)


class RegressionForestTree(ForestTreeMixin.Regression, ForestTreeMixin, RegressionTree):
    """
    """
    pass

class ClassificationForestTree(ForestTreeMixin.Classification, ForestTreeMixin, ClassificationTree):
    """
    """
    pass

# OOB?
class RandomForest(model.Committee):
    """
    """
    
    def __init__(self, numTrees, treeFactory=None):
        model.Committee.__init__()
        while numTrees > 0:
            self.addModel( treeFactory() )
            numTrees -= 1
    
    def fit(self, data):
        """
        """
        for t in self.trees:
            # bootstrap sample
            samples = [random.choice(data) for i in range(len(data))]
            
            # grow tree on bootstrap sample
            t.fit(samples)
        return self


