
from numpy.random import uniform

# Metropolis Sampling
def _mhSamplerN(qSampler, qDist, pDist, numSamples, f, x0=0.0):
    """
    
    :param qSampler:        Sampling function of the proposal distribution.
    :type qSampler:         a -> a
    :param qDist:           Probability density function of the proposa
                            distribution.
    :type qDist:            (Num b) => a -> a -> b
    :param pDist:           Probability density function of the target
                            distribution.
    :type pDist:            (Num b) => a -> b
    :param numSamples:      Number of samples
    :type numSamples:       int
    :param f:               
    :type f:                (Num b) => *pDist* -> *qDist* -> a -> a -> b
    :param x0:              Initial value
    :type x0:               a
    :returns:               *numSamples* samples, distributed according to the
                            target distribution *pDist*.
    
    """
    g = _mhSampler(qSampler, qDist, pDist, f, x0)
    return [g.next() for i in range(numSamples)]

def _mhSampler(qSampler, qDist, pDist, f, last=0.0):
    """
    
    :param qSampler:        Sampling function of the proposal distribution.
    :type qSampler:         a -> a
    :param qDist:           Probability density function of the proposa
                            distribution.
    :type qDist:            (Num b) => a -> a -> b
    :param pDist:           Probability density function of the target
                            distribution.
    :type pDist:            (Num b) => a -> b
    :param f:               
    :type f:                
    :param last:            
    :type last:             float
    :returns:               *numSamples* samples, distributed according to the
                            target distribution *pDist*.
    
    """
    while True:
        u = uniform()
        xs = qSampler( last )
        
        # Metropolis-Hastings
        #A = min(1, pDist(xs) * qDist(last, xs) / ( pDist(last) * qDist(xs, last) ))
        
        # Metropolis; Symmetric proposal, q(x* | x(t)) = q(x(t) | x*)
        #A = min(1, pDist(xs) / pDist( last ))
        
        A = min(1, f(pDist, qDist, xs, last))
        
        if u < A:
            last = xs
            yield xs
        else:
            yield last


def metropolisSamplerN(qSampler, qDist, pDist, numSamples, x0=0.0):
    """Draw samples from a distribution.

    :param qSampler:        Sampling function of the proposal distribution.
    :type qSampler:         a -> a
    :param pDist:           Probability density function of the target
                            distribution.
    :type pDist:            (Num b) => a -> b
    :param numSamples:      Number of samples
    :type numSamples:       int
    :param x0:              Initial value
    :type x0:               a
    :returns:               *numSamples* samples, distributed according to the
                            target distribution *pDist*.
    
    """
    f = lambda pDist, qDist, xs, last: pDist(xs) / pDist(last)
    return _mhSamplerN(qSampler, qDist, pDist, numSamples, f, x0)


def metropolisHastingsSamplerN(qSampler, qDist, pDist, numSamples, x0=0.0):
    """Draw samples from a distribution.
    
    :param qSampler:        Sampling function of the proposal distribution.
    :type qSampler:         a -> a
    :param qDist:           Probability density function of the proposa
                            distribution.
    :type qDist:            (Num b) => a -> a -> b
    :param pDist:           Probability density function of the target
                            distribution.
    :type pDist:            (Num b) => a -> b
    :param numSamples:      Number of samples
    :type numSamples:       int
    :param x0:              Initial value
    :type x0:               a
    :returns:               Generator object for samples, distributed according
                            to the target distribution *pDist*.
    
    """
    f = lambda pDist, qDist, xs, last: (pDist(xs) * qDist(last, xs)) / (pDist(last) * qDist(xs, last))
    return _mhSamplerN(qSampler, qDist, pDist, numSamples, f, x0)


def metropolisSampler(qSampler, qDist, pDist, x0=0.0):
    """Draw samples from a distribution.

    :param qSampler:        Sampling function of the proposal distribution.
    :type qSampler:         a -> a
    :param pDist:           Probability density function of the target
                            distribution.
    :type pDist:            (Num b) => a -> b
    :param x0:              Initial value
    :type x0:               a
    :returns:               *numSamples* samples, distributed according to the
                            target distribution *pDist*.
    
    """
    f = lambda pDist, qDist, xs, last: pDist(xs) / pDist(last)
    return _mhSampler(qSampler, qDist, pDist, f, x0)


def metropolisHastingsSampler(qSampler, qDist, pDist, x0=0.0):
    """Draw samples from a distribution.
    
    :param qSampler:        Sampling function of the proposal distribution.
    :type qSampler:         a -> a
    :param qDist:           Probability density function of the proposal
                            distribution.
    :type qDist:            (Num b) => a -> a -> b
    :param pDist:           Probability density function of the target
                            distribution.
    :type pDist:            (Num b) => a -> b
    :param x0:              Initial value
    :type x0:               a
    :returns:               Generator object for samples, distributed according
                            to the target distribution *pDist*.
    
    """
    f = lambda pDist, qDist, xs, last: (pDist(xs) * qDist(last, xs)) / (pDist(last) * qDist(xs, last))
    return _mhSampler(qSampler, qDist, pDist, f, x0)


# TODO: Gibbs sampling for mixtures


